/*
This example was put together to show how to call a recursive function from your main function in order to go through an entire DynamoDB table looking for matching rows to a scan request.  Due to DynamoDB design limitations, a maximum of 1MB of data can be examined at a time, hence the need for recursive function.
*/

/*
AWS DYNAMO
*/
var AWS = require("aws-sdk");
var docClient = new AWS.DynamoDB.DocumentClient();
var params = {};

/*
FUNCTION THAT IS CALLED RECURSIVELY TO CRAWL AN ENTIRE TABLE LOOKING FOR SEARCH MATCHES UNTIL THE ENTIRE TABLE HAS BEEN SEARCHED
*/
var repeater = function(params, callback, items) {
	docClient.scan(params, function(err, data) {
		if (err) {
			/* DynamoDB error */
			noitems = [];
			callback(err, noitems);
		} else {
			/* Process results, if any */
			var tmp_SearchValue = "";
			tmp_LastEvaluatedKey = {};
			/* JS not happy unless you make a copy of "items" and then reference the copy; "items" will be seen as undefined */
			tmp_items = [];
			if (items) { tmp_items = items; }
			data.Items.forEach(function(row) {
				if (row.SearchValue) {
					tmp_SearchValue = row.SearchValue;
					tmp_items[tmp_items.length] = {
						"title": tmp_SearchValue
					};
				}
			});
			/*
			  Determine if we have follow-up calls to make (a 1MB limit has been reached) or the end of the table has been reached.
			  Note: There is no intuitive way with which to instruct Dynamo that you want a set of X matching results returned at a time (whether or not there may be more results).  "Limit" does not work the way you may think...definitely NOT like "Limit" in SQL works.
			        Instead you must go the archaic route and crawl the ENTIRE table to get matching results, cache the ENTIRE RESULT SET and then slice it up and ship it out in chunks of X results.
			*/
			if (typeof data.LastEvaluatedKey !== "undefined" && typeof data.LastEvaluatedKey !== null) {
				/* We've reached a 1MB design limitation, prepare for another recursive call to go through the next 1MB */
				/* Get the last row the query examined */
				tmp_LastEvaluatedKey = data.LastEvaluatedKey;
				/* Add "ExclusiveStartKey" to the params query */
				params.ExclusiveStartKey = tmp_LastEvaluatedKey;
				/* Perform the next 1MB crawl */
				repeater(params, callback, tmp_items);
			}
			else {
				/* No more rows in the table to examine, return */
				callback(err, tmp_items);
			}
		}
	});
}

/*
DEFAULT MAIN FUNCTION CALL
*/
exports.handler = function(event, context) {
	/*
	INPUT PARAMETERS:
	SearchValue =	the keyword(s) that the user has entered to search for
	SearchType =	the column to search
	INPUT EXAMPLE:
		{
		"SearchType": "apple"
		"SearchValue": "honey"
		}
	*/

	/* Default input parameter values, if none are specified */
	var SearchValue = "honey";

	/* Gather inputs */
	if (event.SearchValue) { SearchValue = event.SearchValue.trim(); }

	/* Define recursive query to utilize */
	/*
	  In this example, the table "testSearch" is comprised of a partition key "SearchType" (string like apple for fruit type) and a range key "SearchValue" (string like honeycrisp for fruit name)
	*/
	params = {
		TableName: "testSearch",
		ExclusiveStartKey: null,
		ProjectionExpression: "SearchType, SearchValue",
		FilterExpression: "SearchType = :SearchType AND contains (SearchValue, :datainSearchValue)",
		ExpressionAttributeValues: {
			":datainSearchValue": SearchValue
		}
	};

	/* Define the callback structure as the result of making recursive calls to the table until all rows have been crawled (due to paltry 1MB limitation) */
	var resultsCollection = function(err, data) {
		if (err) {
			/* Some type of error was encountered from the use of the function "repeater" */
			context.done(
				null,
				{
				"Error": JSON.stringify(err, null, 2)
				}
			);
		}
		else {
			/* The table rows have been crawled.  Let's see if we have search results to put into the array "SearchResults" */
			SearchResults = [];
			data.forEach(function(row) {
				if (row.title) {
					tmp_title = row.title;
					SearchResults[SearchResults.length] = {
						"title": row.title.toString()
					};
				}
			});
			/* If we have any matching results from crawling the entire table, return it */
			if (SearchResults.length > 0) {
				/* The 90's are calling...if you want to, say, provide a set of X search results at a time to a user, cache this nightmare set of matching results (which may contain thousands of matching results) somewhere specifically for the user, and then serve the user X search results at a time from that cache */
				context.done(
					null,
					{
					"NumberOfResults": SearchResults.length,
					"SearchResults": SearchResults
					}
				);
			}
			else {
				/* No search matches were found after crawling the entire table */
				context.done(
					null,
					{
					"NoResults": "No results were found"
					}
				);
			}
		}
	};
	/* Begin the recursive process to crawl the table in 1MB chunks.  Once the entire table has been crawled, or an error occurs, then the function above "resultsCollection" will be invoked */
	repeater(params, resultsCollection);
};